require '../../spec_helper'

describe "Freakset", ->

  describe 'instance', ->

    beforeEach ->
      @freakset = Freakset.build(store: { name: "value" })


    describe '#new', ->

      it 'should init stack and store with its correct empty values', ->
        flow = Freakset.build()
        expect(flow.stack).toEqual([])
        expect(flow.store).toEqual({})

      it 'should override initial values', ->
        flow = Freakset.build(stack: "stack", store: "store")
        expect(flow.stack).toEqual("stack")
        expect(flow.store).toEqual("store")


    describe '#isRunning', ->
      it 'should return false if the set has finished', ->
        @freakset.stack = []
        @freakset.commit()
        expect(@freakset.isRunning()).toBe(false)

      it 'should return true if the set is still running', ->
        @freakset.stack = [ "sa" ]
        @freakset.commit()
        expect(@freakset.isRunning()).toBe(true)

      it 'should return true if at least one innerFreakset is still running', ->
        inner = Freakset.build()
        inner.running = yes
        @freakset.innerFreaksets.push(inner)
        expect(@freakset.isRunning()).toBe(true)

      it 'should return true a inner-innerFreakset is still running', ->
        innerInner = Freakset.build()
        innerInner.running = yes
        inner = Freakset.build()
        inner.running = no
        inner.innerFreaksets.push(innerInner)
        @freakset.innerFreaksets.push(inner)
        expect(@freakset.isRunning()).toBe(true)


    describe '#root', ->

      it 'should return self if there is no outer Freakset', ->
        expect(@freakset.root()).toEqual(@freakset)

      it 'should return the most outer Freakset', ->
        inner = Freakset.build(outerFreakset: @freakset)
        innerInner = Freakset.build(outerFreakset: inner)
        expect(innerInner.root()).toEqual(@freakset)


    describe '#step', ->

      it 'should append a actor definition to the stack', ->
        @freakset.step "spec:scope", ->
        expect(@freakset.stack.length).toBe(1)

      it 'should attach type, scope and callback to the stack', ->
        fn = ->
        @freakset.step "spec:step", fn
        expect(@freakset.stack[0]).toEqual({ type: "step", scope: "spec:step", fn: fn })


    describe '#group', ->

      it 'should create a inner Freakset with an empty stack, passing store and the current Freakset (as outer)', ->
        spyOn(Freakset, 'build')
        @freakset.group "scope:group", ->
        expect(Freakset.build).toHaveBeenCalledWith(store: @freakset.store, outerFreakset: @freakset, runInParallel: no)

      it 'should attach the actor definition to the stack', ->
        fn = ->
        spyOn(Freakset, 'build').andReturn("innerFreakset")
        @freakset.group "scope:group", fn
        expect(@freakset.stack[0]).toEqual({ type: "group", scope: "scope:group", fn: fn, innerFreakset: "innerFreakset", parallel: false })

      it 'should pass the parallel option to the actor definition', ->
        fn = ->
        spyOn(Freakset, 'build').andReturn("innerFreakset")
        @freakset.group "scope:group", { parallel: yes }, fn
        expect(@freakset.stack[0]).toEqual({ type: "group", scope: "scope:group", fn: fn, innerFreakset: "innerFreakset", parallel: true })

      it 'should invoke the callback and pass the innerFreakset as argument', ->
        fn = jasmine.createSpy()
        spyOn(Freakset, 'build').andReturn("innerFreakset")
        @freakset.group "scope:group", fn
        expect(fn).toHaveBeenCalledWith('innerFreakset')


    describe '#rescue', ->

      it 'should create a inner Freakset with an empty stack, passing store and the current Freakset (as outer)', ->
        spyOn(Freakset, 'build')
        @freakset.rescue "scope:rescue", undefined, ->
        expect(Freakset.build).toHaveBeenCalledWith(store: @freakset.store, outerFreakset: @freakset)

      it 'should attach the actor definition to the stack', ->
        fn = ->
        spyOn(Freakset, 'build').andReturn("innerFreakset")
        @freakset.rescue "spec:rescue", undefined, fn
        expect(@freakset.stack[0]).toEqual({ type: "rescue", scope: "spec:rescue", fn: fn, innerFreakset: "innerFreakset" })

      it 'should attach the errFn callback to the actor definition', ->
        fn = ->
        errFn = ->
        spyOn(Freakset, 'build').andReturn("innerFreakset")
        @freakset.rescue "spec:rescue", errFn, fn
        expect(@freakset.stack[0]['errFn']).toEqual(errFn)


      it 'should invoke the callback and pass the innerFreakset as argument', ->
        fn = jasmine.createSpy()
        spyOn(Freakset, 'build').andReturn("innerFreakset")
        @freakset.rescue "spec:rescue", undefined, fn
        expect(fn).toHaveBeenCalledWith('innerFreakset')

      it 'should keep references of innerFreaksets', ->
        fn = ->
        spyOn(Freakset, 'build').andReturn("innerFreakset")
        @freakset.rescue "spec:rescue", undefined, fn
        expect(@freakset.innerFreaksets.length).toBe(1)
        expect(@freakset.innerFreaksets[0]).toEqual('innerFreakset')

      it 'should return false all inner(inner(inner)) sets have finished', ->
        innerInner = Freakset.build()
        innerInner.running = no
        inner = Freakset.build()
        inner.running = no
        inner.innerFreaksets.push(innerInner)
        @freakset.innerFreaksets.push(inner)
        expect(@freakset.isRunning()).toBe(false)

      it 'should return the new (inner) Freakset', ->
        spyOn(Freakset, 'build').andReturn("innerFreakset")
        expect(@freakset.rescue "spec:rescue", undefined, ->).toEqual('innerFreakset')


    describe '#guard', ->

      it 'should attach the actor definition to the stack', ->
        fn = ->
        @freakset.guard "spec:guard", fn
        expect(@freakset.stack[0]).toEqual({ type: "guard", scope: "spec:guard", fn: fn })


    describe '#next', ->

      it 'should mark the set as running', ->
        @freakset.stack = [ "step" ]
        @freakset.next()
        expect(@freakset.isRunning()).toBe(true)

      it 'should get and remove the first actor of the stack', ->
        @freakset.stack = [ "test" ]
        @freakset.next()
        expect(@freakset.stack.length).toBe(0)

      it 'should return true if the stack is empty', ->
        @freakset.stack = []
        expect(@freakset.next()).toBe(true)

      it 'should #finish the freakset if the stack is empty', ->
        spyOn(@freakset, "finish")
        @freakset.stack = []
        @freakset.next()
        expect(@freakset.finish).toHaveBeenCalled()


    describe '#commit', ->

      it 'should call #next on the current Freakset', ->
        spyOn(@freakset, 'next')
        @freakset.commit()
        expect(@freakset.next).toHaveBeenCalled()

      it 'should decrease the parallel step count if the Freakset is running in parallel', ->
        @freakset.next = jasmine.createSpy()
        @freakset.runInParallel = yes
        @freakset.parallelStepCount = 2
        @freakset.commit()
        expect(@freakset.parallelStepCount).toBe(1)


    describe '#runRescueActor', ->

      it 'should build and run the inner freakset', ->
        spyOn(@freakset, 'buildAndRunFreakset')
        innerFreakset = {}
        innerFreakset['stack'] = [ 'innerStep' ]
        @freakset.runRescueActor({ type: "rescue", innerFreakset: innerFreakset }) 
        expect(@freakset.buildAndRunFreakset).toHaveBeenCalledWith(stack: innerFreakset.stack, store: @freakset.store, outerFreakset: @freakset)

      describe 'catching errors', ->
        beforeEach ->
          @errFn = jasmine.createSpy()
          @step = { type: "step", fn: -> ( throw 42 ) }
          @innerFreakset = {}
          @innerFreakset['stack'] = [ @step ]
          @actor = { type: "rescue", innerFreakset: @innerFreakset, errFn: @errFn }

        it 'should not throw an error raised by an inner step', ->
          @freakset.runRescueActor(@actor)

        it 'should call the errFn and pass the err, the current Freakset and step', ->
          @freakset.runRescueActor(@actor)
          expect(@errFn).toHaveBeenCalledWith(42, @freakset, @innerFreakset)


    describe '#runGuardActor', ->

      it 'should run the actors callback', ->
        fn = jasmine.createSpy()
        @freakset.runGuardActor({ type: "guard", fn: fn })
        expect(fn).toHaveBeenCalledWith(@freakset)

      it 'should #finish the current Freakset if the callback return false', ->
        fn = -> ( false )
        spyOn(@freakset, "finish")
        @freakset.runGuardActor({ type: "guard", fn: fn })
        expect(@freakset.finish).toHaveBeenCalled()


    describe '#runStepActor', ->

      it 'should run the actor in the context of this', ->
        fn = jasmine.createSpy()
        @freakset.runStepActor({ type: "step", fn: fn })
        expect(fn).toHaveBeenCalledWith(@freakset)


    describe '#compile', ->

      it 'should build and commit the root freakset', ->
        spyOn(@freakset, "buildAndRunFreakset")
        @freakset.compile()
        expect(@freakset.buildAndRunFreakset).toHaveBeenCalledWith(stack: @freakset.stack, store: {})

      it 'should build a store from the passed object', ->
        spyOn(@freakset, "buildAndRunFreakset")
        @freakset.compile({ test: "foo" })
        expect(@freakset.buildAndRunFreakset).toHaveBeenCalledWith(stack: @freakset.stack, store: { test: "foo" })

      it 'should attach a resolvedCallback if passed', ->
        fn = ->
        @freakset.compile {}, fn
        expect(@freakset.resolvedCallback).toBe(fn)

      it 'should return the newly build Freakset', ->
        spyOn(@freakset, "buildAndRunFreakset").andReturn('newFreakset')
        expect(@freakset.compile({})).toEqual('newFreakset')


    describe '#resolveRoot', ->

      it 'should invoke the resolvedCallback (if defined) after the set has finished', ->
        fn = ->
        spyOn(fn, 'call')
        @freakset.resolvedCallback = fn
        @freakset.running = no
        @freakset.resolveRoot()
        expect(fn.call).toHaveBeenCalled()


    describe '#finish', ->

      it 'should mark the freakset as not running anymore', ->
        @freakset.stack = []
        @freakset.finish()
        expect(@freakset.running).toBe(false)

      it 'should call commit on the outerFreakset', ->
        @freakset.stack = []
        @freakset.outerFreakset = Freakset.build()
        spyOn(@freakset.outerFreakset, "commit")
        @freakset.finish()
        expect(@freakset.outerFreakset.commit).toHaveBeenCalled()

      it 'should ask its root to resolve', ->
        spyOn(@freakset, "resolveRoot")
        @freakset.stack = []
        @freakset.finish()
        expect(@freakset.resolveRoot).toHaveBeenCalled()


    describe 'internal helper', ->


      describe '#buildAndRunFreakset', ->

        it 'should build new Freakset and commit it afterwards', ->
          set = { next: -> }
          spyOn(Freakset, 'build').andReturn(set)
          spyOn(set, 'next')
          @freakset.buildAndRunFreakset( [] )
          expect(set.next).toHaveBeenCalled()

        it 'should return the new Freakset', ->
          newFreakset = Freakset.build()
          spyOn(Freakset, 'build').andReturn(newFreakset)
          expect(@freakset.buildAndRunFreakset([])).toEqual(newFreakset)


        it 'should pass a copy of the stack to the new Freakset', ->
          stack = []
          spyOn(stack, "slice")
          @freakset.buildAndRunFreakset(stack: stack)
          expect(stack.slice).toHaveBeenCalled()

        it 'should pass the store and outerFreakset to the new one', ->
          spyOn(Freakset, 'build').andReturn { next: ->}
          @freakset.buildAndRunFreakset(stack: ["stack"], store: { name: 'Iam' }, outerFreakset: { test: 'Iam' } )
          expect(Freakset.build).toHaveBeenCalledWith( stack: ["stack"], store: { name: 'Iam' }, outerFreakset: { test: 'Iam' } )


      describe '#pushToStack', ->

        it 'should add the passed object to the stack', ->
          stack = []
          @freakset.pushToStack stack, { foo: 'bar', peter: 'fox' }
          expect(stack).toEqual( [ { foo: 'bar', peter: 'fox' } ])

        it 'should add the passed object to the top of the stack if specified (opts.inFront=yes)', ->
          stack = [ { existing: "step" }]
          @freakset.pushToStack stack, { newer: 'step' }, { inFront: yes}
          expect(stack).toEqual([ {newer: 'step'}, {existing: "step"}  ])

        it 'should add the object to the top if the current Freakset is running (add Steps on the Fly)', ->
          stack = [ { existing: "step" }]
          @freakset.running = yes
          @freakset.pushToStack stack, { newer: 'step' }
          expect(stack).toEqual([ {newer: 'step'}, {existing: "step"}  ])
